from collections import defaultdict
from typing import List, Optional

import numpy as np

from pymilvus.model.base import BaseEmbeddingFunction
from pymilvus.model.utils import import_openai


class OpenAIEmbeddingFunction(BaseEmbeddingFunction):
    def __init__(
        self,
        model_name: str = "text-embedding-ada-002",
        api_key: Optional[str] = None,
        base_url: Optional[str] = None,
        dimensions: Optional[int] = None,
        **kwargs,
    ):
        import_openai()
        from openai import OpenAI

        self._openai_model_meta_info = defaultdict(dict)
        self._openai_model_meta_info["text-embedding-3-small"]["dim"] = 1536
        self._openai_model_meta_info["text-embedding-3-large"]["dim"] = 3072
        self._openai_model_meta_info["text-embedding-ada-002"]["dim"] = 1536

        self._model_config = dict({"api_key": api_key, "base_url": base_url}, **kwargs)
        additional_encode_config = {}
        if dimensions is not None:
            additional_encode_config = {"dimensions": dimensions}
            self._openai_model_meta_info[model_name]["dim"] = dimensions

        self._encode_config = {"model": model_name, **additional_encode_config}
        self.model_name = model_name
        self.client = OpenAI(**self._model_config)

    def encode_queries(self, queries: List[str]) -> List[np.array]:
        return self._encode(queries)

    def encode_documents(self, documents: List[str]) -> List[np.array]:
        return self._encode(documents)

    @property
    def dim(self):
        return self._openai_model_meta_info[self.model_name]["dim"]

    def __call__(self, texts: List[str]) -> List[np.array]:
        return self._encode(texts)

    def _encode_query(self, query: str) -> np.array:
        return self._encode(query)[0]

    def _encode_document(self, document: str) -> np.array:
        return self._encode(document)[0]

    def _call_openai_api(self, texts: List[str]):
        results = self.client.embeddings.create(input=texts, **self._encode_config).data
        return [np.array(data.embedding) for data in results]

    def _encode(self, texts: List[str]):
        return self._call_openai_api(texts)



import logging
import re
import string
from importlib.util import find_spec
from pathlib import Path
from typing import Any, Dict, List, Match, Optional, Type
from pymilvus.model.utils import import_nltk, import_jieba, import_mecab, import_konlpy, import_unidic_lite


logger = logging.getLogger(__name__)
logger.setLevel(logging.DEBUG)

_class_registry = {}


def register_class(register_as: str):
    def decorator(cls: Type[Any]):
        _class_registry[register_as] = cls
        return cls

    return decorator


class Preprocessor:
    def apply(self, text: str):
        error_message = "Each preprocessor must implement its 'apply' method."
        raise NotImplementedError(error_message)


@register_class("CharacterfilterPreprocessor")
class CharacterfilterPreprocessor:
    def __init__(self, chars_to_replace: str):
        self.replacement_table = str.maketrans({char: " " for char in chars_to_replace})

    def apply(self, text: str):
        return text.translate(self.replacement_table)


@register_class("ReplacePreprocessor")
class ReplacePreprocessor:
    def __init__(self, replacement_mapping: Dict[str, str]):
        self.replacement_mapping = replacement_mapping
        self.pattern = re.compile("|".join(map(re.escape, replacement_mapping.keys())))

    def _replacement_function(self, match: Match):
        return self.replacement_mapping[match.group(0)]

    def apply(self, text: str):
        return self.pattern.sub(self._replacement_function, text)


@register_class("StandardTokenizer")
class StandardTokenizer:
    def __init__(self):
        import_nltk()
        import nltk
        from nltk import word_tokenize
        try:
           word_tokenize("this is a simple test.") 
        except LookupError:
            nltk.download("punkt_tab")
    def tokenize(self, text: str):
        from nltk import word_tokenize
        return word_tokenize(text)


class TextFilter:
    def apply(self, tokens: List[str]):
        error_message = "Each filter must implement the 'apply' method."
        raise NotImplementedError(error_message)


@register_class("LowercaseFilter")
class LowercaseFilter(TextFilter):
    def apply(self, tokens: List[str]):
        return [token.lower() for token in tokens]


@register_class("StopwordFilter")
class StopwordFilter(TextFilter):
    def __init__(self, language: str = "english", stopword_list: Optional[List[str]] = None):
        import_nltk()
        import nltk
        from nltk.corpus import stopwords
        try:
            nltk.corpus.stopwords.words(language)
        except LookupError:
            nltk.download("stopwords")

        if stopword_list is None:
            stopword_list = []
        self.stopwords = set(stopwords.words(language) + stopword_list)

    def apply(self, tokens: List[str]):
        return [token for token in tokens if token not in self.stopwords]


@register_class("PunctuationFilter")
class PunctuationFilter(TextFilter):
    def __init__(self, extras: str = ""):
        self.punctuation = set(string.punctuation + extras)

    def apply(self, tokens: List[str]):
        return [token for token in tokens if token not in self.punctuation]


@register_class("StemmingFilter")
class StemmingFilter(TextFilter):
    def __init__(self, language: str = "english"):
        import_nltk()
        from nltk.stem.snowball import SnowballStemmer
        self.stemmer = SnowballStemmer(language)

    def apply(self, tokens: List[str]):
        return [self.stemmer.stem(token) for token in tokens]


class Tokenizer:
    def tokenize(self, text: str):
        error_message = "Each tokenizer must implement its 'tokenize' method."
        raise NotImplementedError(error_message)


@register_class("JiebaTokenizer")
class JiebaTokenizer(Tokenizer):
    def __init__(self):
        import_jieba()
        if find_spec("jieba") is None:
            error_message = "jieba is required for JiebaTokenizer but is not installed. Please install it using 'pip install jieba'."
            logger.error(error_message)
            raise ImportError(error_message)

    def tokenize(self, text: str):
        import jieba

        return jieba.lcut(text)


@register_class("MecabTokenizer")
class MecabTokenizer(Tokenizer):
    def __init__(self):
        import_unidic_lite()
        import_mecab()
        if find_spec("MeCab") is None:
            error_message = "MeCab is required for MecabTokenizer but is not installed. Please install it using 'pip install mecab-python3'."
            logger.error(error_message)
            raise ImportError(error_message)

    def tokenize(self, text: str):
        import MeCab

        wakati = MeCab.Tagger("-Owakati")
        return wakati.parse(text).split()


@register_class("KonlpyTokenizer")
class KonlpyTokenizer(Tokenizer):
    def __init__(self):
        import_konlpy()
        if find_spec("konlpy") is None:
            error_message = "konlpy is required for KonlpyTokenizer but is not installed. Please install it using 'pip install konlpy'."
            logger.error(error_message)
            raise ImportError(error_message)

    def tokenize(self, text: str):
        from konlpy.tag import Kkma

        return Kkma().nouns(text)


class Analyzer:
    def __init__(
        self,
        name: str,
        tokenizer: Tokenizer,
        preprocessors: Optional[List[Preprocessor]] = None,
        filters: Optional[List[TextFilter]] = None,
    ):
        self.name = name
        self.tokenizer = tokenizer
        self.preprocessors = preprocessors
        self.filters = filters

    def __call__(self, text: str):
        for preprocessor in self.preprocessors:
            text = preprocessor.apply(text)
        tokens = self.tokenizer.tokenize(text)
        for _filter in self.filters:
            tokens = _filter.apply(tokens)
        return tokens


def build_default_analyzer(language: str = "en"):
    default_config_path = Path(__file__).parent / "lang.yaml"
    return build_analyzer_from_yaml(default_config_path, language)


def build_analyzer_from_yaml(filepath: str, name: str):
    import yaml

    with Path(filepath).open(encoding="utf-8") as file:
        config = yaml.safe_load(file)

    lang_config = config.get(name)
    if not lang_config:
        error_message = f"No configuration found {name}"
        raise ValueError(error_message)

    tokenizer_class_type = _class_registry[lang_config["tokenizer"]["class"]]
    tokenizer_params = lang_config["tokenizer"]["params"]

    tokenizer = tokenizer_class_type(**tokenizer_params)
    preprocessors = []
    filters = []
    if "preprocessors" in lang_config:
        preprocessors = [
            _class_registry[filter_config["class"]](**filter_config["params"])
            for filter_config in lang_config["preprocessors"]
        ]
    if "filters" in lang_config:
        filters = [
            _class_registry[filter_config["class"]](**filter_config["params"])
            for filter_config in lang_config["filters"]
        ]

    return Analyzer(name=name, tokenizer=tokenizer, preprocessors=preprocessors, filters=filters)



import numpy as np
from typing import List, Union
from pathlib import Path
from pymilvus.model.base import BaseEmbeddingFunction
from pymilvus.model.utils import import_model2vec


class Model2VecEmbeddingFunction(BaseEmbeddingFunction):
    def __init__(self, model_source: Union[str, Path] = "minishlab/potion-base-8M", **kwargs):
        """
        Initialize the Model2VecEmbeddingFunction, which loads a Model2Vec model either from the Hugging Face Hub or from a local directory.
        Defaults to use the "minishlab/potion-base-8M" model and load from Hugging Face.

        Parameters:
            model_source (Union[str, Path]):
                - If a string is provided and it does not correspond to an existing local directory,
                  it is interpreted as a Hugging Face model identifier (e.g., "minishlab/potion-base-8M").
                - If the provided string (or Path) corresponds to an existing directory, the model is loaded locally.
            **kwargs:
                - Additional keyword arguments that will be passed to the StaticModel.from_pretrained method
                  when loading a remote model from the Hugging Face Hub, including parameters such as
                  huggingface authentication tokens.
        """
        import_model2vec()
        from model2vec import StaticModel

        self.model_source = model_source
        model_path = Path(model_source)

        if model_path.exists() and model_path.is_dir():
            self.model = StaticModel.load_local(model_path)
        else:
            self.model = StaticModel.from_pretrained(model_source, **kwargs)

        dummy_embedding = self.model.encode(["dummy"])
        self._dim = dummy_embedding[0].shape[0]

    @property
    def dim(self) -> int:
        return self._dim

    def encode_queries(self, queries: List[str]) -> List[np.array]:
        return self._encode(queries)

    def encode_documents(self, documents: List[str]) -> List[np.array]:
        return self._encode(documents)

    def _encode_query(self, query: str) -> np.array:
        return self._encode([query])[0]

    def _encode_document(self, document: str) -> np.array:
        return self._encode([document])[0]

    def __call__(self, texts: List[str]) -> List[np.array]:
        return self._encode(texts)

    def _encode(self, texts: List[str]) -> List[np.array]:
        embeddings = self.model.encode(texts)
        return [embeddings[i] for i in range(embeddings.shape[0])]


from typing import List

import requests

from pymilvus.model.base import BaseRerankFunction, RerankResult


class TEIRerankFunction(BaseRerankFunction):
    def __init__(self, api_url: str):
        self.api_url = api_url + "/rerank"
        self._session = requests.Session()

    def __call__(self, query: str, documents: List[str], top_k: int = 5) -> List[RerankResult]:
        resp = self._session.post(  # type: ignore[assignment]
            self.api_url,
            json={
                "query": query,
                "return_text": True,
                "texts": documents,
            },
        ).json()
        if "error" in resp:
            raise RuntimeError(resp["error"])

        results = []
        for res in resp[:5]:
            results.append(RerankResult(text=res["text"], score=res["score"], index=res["index"]))
        return results



from typing import List, Optional

import numpy as np
import requests

from pymilvus.model.base import BaseEmbeddingFunction


class TEIEmbeddingFunction(BaseEmbeddingFunction):
    def __init__(
        self,
        api_url: str,
        dimensions: Optional[int] = None,
    ):
        self.api_url = api_url + "/v1/embeddings"
        self._session = requests.Session()
        self._dim = dimensions

    @property
    def dim(self):
        if self._dim is None:
            # This works by sending a dummy message to the API to retrieve the vector dimension,
            # as the original API does not directly provide this information
            self._dim = self._call_api(["get dim"])[0].shape[0]
        return self._dim

    def encode_queries(self, queries: List[str]) -> List[np.array]:
        return self._call_api(queries)

    def encode_documents(self, documents: List[str]) -> List[np.array]:
        return self._call_api(documents)

    def __call__(self, texts: List[str]) -> List[np.array]:
        return self._call_api(texts)

    def _call_api(self, texts: List[str]):
        data = {"input": texts}
        resp = self._session.post(  # type: ignore[assignment]
            self.api_url,
            json=data,
        ).json()
        if "data" not in resp:
            raise RuntimeError(resp["message"])

        embeddings = resp["data"]

        # Sort resulting embeddings by index
        sorted_embeddings = sorted(embeddings, key=lambda e: e["index"])  # type: ignore[no-any-return]
        return [np.array(result["embedding"]) for result in sorted_embeddings]



from typing import List, Optional
import struct
from collections import defaultdict
import numpy as np

from pymilvus.model.base import BaseEmbeddingFunction
from pymilvus.model.utils import import_cohere



class CohereEmbeddingFunction(BaseEmbeddingFunction):
    def __init__(self,
                 model_name: str = "embed-english-light-v3.0",
                 api_key: Optional[str] = None,
                 input_type: str = "search_document",
                 embedding_types: Optional[List[str]] = None,
                 truncate: Optional[str] = None,
                 **kwargs):
        self.model_name = model_name
        self.input_type = input_type
        self.embedding_types = embedding_types
        self.truncate = truncate

        import_cohere()
        import cohere

        if isinstance(embedding_types, list):
            if len(embedding_types) > 1:
                raise ValueError("Only one embedding type can be specified using current PyMilvus model library.")
            elif embedding_types[0] == "int8" or embedding_types[0] == "uint8":
                raise ValueError("Currently int8 or uint8 is not supported with PyMilvus model library.")
            else:
                pass
    
        self.client = cohere.Client(api_key, **kwargs)
        self._cohereai_model_meta_info = defaultdict(dict)
        self._cohereai_model_meta_info["embed-english-v3.0"]["dim"] = 1024
        self._cohereai_model_meta_info["embed-english-light-v3.0"]["dim"] = 384
        self._cohereai_model_meta_info["embed-english-v2.0"]["dim"] = 4096
        self._cohereai_model_meta_info["embed-english-light-v2.0"]["dim"] = 1024
        self._cohereai_model_meta_info["embed-multilingual-v3.0"]["dim"] = 1024
        self._cohereai_model_meta_info["embed-multilingual-light-v3.0"]["dim"] = 384
        self._cohereai_model_meta_info["embed-multilingual-v2.0"]["dim"] = 768

    def _call_cohere_api(self, texts: List[str], input_type: str) -> List[np.array]:
        embeddings = self.client.embed(
            texts=texts,
            model=self.model_name,
            input_type=input_type,
            embedding_types=self.embedding_types,
            truncate=self.truncate
        ).embeddings
        if self.embedding_types is None:
            results = [np.array(data, dtype=np.float32) for data in embeddings]
        else:
            results = getattr(embeddings, self.embedding_types[0], None)
            if self.embedding_types[0] == "binary":
                results = [struct.pack('b' * len(int8_vector), *int8_vector) for int8_vector in results] 
            elif self.embedding_types[0] == "ubinary":
                results = [struct.pack('B' * len(uint8_vector), *uint8_vector) for uint8_vector in results] 
            elif self.embedding_types[0] == "float":
                results = [np.array(result, dtype=np.float32) for result in results]
            else:
                pass
        return results

    def encode_documents(self, documents: List[str]) -> List[np.array]:
        return self._call_cohere_api(documents, input_type="search_document")

    def encode_queries(self, queries: List[str]) -> List[np.array]:
        return self._call_cohere_api(queries, input_type="search_query")

    def __call__(self, texts: List[str]) -> List[np.array]:
        return self._call_cohere_api(texts, self.input_type)

    @property
    def dim(self):
        return self._cohereai_model_meta_info[self.model_name]["dim"]




from pymilvus.model.sparse.bm25.bm25 import BM25EmbeddingFunction
from pymilvus.model.sparse.bm25.tokenizers import Analyzer, build_analyzer_from_yaml, build_default_analyzer

__all__ = [
    "BM25EmbeddingFunction",
    "Analyzer",
    "build_analyzer_from_yaml",
    "build_default_analyzer",
]



from abc import abstractmethod
from typing import List


class BaseEmbeddingFunction:
    @abstractmethod
    def __call__(self, texts: List[str]):
        """ """

    @abstractmethod
    def encode_queries(self, queries: List[str]):
        """ """


class BaseRerankFunction:
    @abstractmethod
    def __call__(self, query: str, documents: List[str], top_k: int):
        """ """


class RerankResult:
    def __init__(self, text: str, score: float, index: int):
        self.text = text
        self.score = score
        self.index = index

    def to_dict(self):
        return {"text": self.text, "score": self.score, "index": self.index}

    def __str__(self):
        return f"RerankResult(text={self.text!r}, score={self.score}, index={self.index})"

    def __repr__(self):
        return f"RerankResult(text={self.text!r}, score={self.score}, index={self.index})"



from typing import Dict
from collections import defaultdict

from pymilvus.model.utils import import_torch, import_scipy, import_transformers

import_torch()
import_scipy()
import_transformers()

import numpy as np
import torch
from transformers import AutoModelForTokenClassification, AutoTokenizer
from transformers.utils import is_torch_npu_available


class _GTEEmbeddidng(torch.nn.Module):
    def __init__(self,
                 model_name: str = None,
                 normalized: bool = True,
                 use_fp16: bool = True,
                 device: str = None
                ):
        super().__init__()
        self.normalized = normalized
        if device:
            self.device = torch.device(device)
        else:
            if torch.cuda.is_available():
                self.device = torch.device("cuda")
            elif torch.backends.mps.is_available():
                self.device = torch.device("mps")
            elif is_torch_npu_available():
                self.device = torch.device("npu")
            else:
                self.device = torch.device("cpu")
                use_fp16 = False
        self.use_fp16 = use_fp16
        self.tokenizer = AutoTokenizer.from_pretrained(model_name)
        self.model = AutoModelForTokenClassification.from_pretrained(
            model_name, trust_remote_code=True, torch_dtype=torch.float16 if self.use_fp16 else None
        )
        self.vocab_size = self.model.config.vocab_size
        self.model.to(self.device)

    def _process_token_weights(self, token_weights: np.ndarray, input_ids: list):
        # conver to dict
        result = defaultdict(int)
        unused_tokens = set([self.tokenizer.cls_token_id, self.tokenizer.eos_token_id, self.tokenizer.pad_token_id,
                             self.tokenizer.unk_token_id])
        for w, idx in zip(token_weights, input_ids):
            if idx not in unused_tokens and w > 0:
                token = int(idx)
                if w > result[token]:
                    result[token] = w
        return result

    @torch.no_grad()
    def encode(self,
               texts: None,
               dimension: int = None,
               max_length: int = 8192,
               batch_size: int = 16,
               return_dense: bool = True,
               return_sparse: bool = False):
        if isinstance(texts, str):
            texts = [texts]
        num_texts = len(texts)
        all_dense_vecs = []
        all_token_weights = []
        for n, i in enumerate(range(0, num_texts, batch_size)):
            batch = texts[i: i + batch_size]
            resulst = self._encode(batch, dimension, max_length, batch_size, return_dense, return_sparse)
            if return_dense:
                all_dense_vecs.append(resulst['dense_embeddings'])
            if return_sparse:
                all_token_weights.extend(resulst['token_weights'])
        all_dense_vecs = torch.cat(all_dense_vecs, dim=0)
        return {
            "dense_embeddings": all_dense_vecs,
            "token_weights": all_token_weights
        }

    @torch.no_grad()
    def _encode(self,
                texts: Dict[str, torch.Tensor] = None,
                dimension: int = None,
                max_length: int = 1024,
                batch_size: int = 16,
                return_dense: bool = True,
                return_sparse: bool = False):

        text_input = self.tokenizer(texts, padding=True, truncation=True, return_tensors='pt', max_length=max_length)
        text_input = {k: v.to(self.model.device) for k,v in text_input.items()}
        model_out = self.model(**text_input, return_dict=True)

        output = {}
        if return_dense:
            dense_vecs = model_out.last_hidden_state[:, 0, :dimension]
            if self.normalized:
                dense_vecs = torch.nn.functional.normalize(dense_vecs, dim=-1)
            output['dense_embeddings'] = dense_vecs
        if return_sparse:
            token_weights = torch.relu(model_out.logits).squeeze(-1)
            token_weights = list(map(self._process_token_weights, token_weights.detach().cpu().numpy().tolist(),
                                                    text_input['input_ids'].cpu().numpy().tolist()))
            output['token_weights'] = token_weights

        return output


import logging
from typing import Dict, List

from scipy.sparse import csr_array
import numpy as np

from pymilvus.model.base import BaseEmbeddingFunction
from pymilvus.model.utils import import_FlagEmbedding, import_datasets
from pymilvus.model.sparse.utils import stack_sparse_embeddings


logger = logging.getLogger(__name__)
logger.setLevel(logging.DEBUG)


class BGEM3EmbeddingFunction(BaseEmbeddingFunction):
    def __init__(
        self,
        model_name: str = "BAAI/bge-m3",
        batch_size: int = 16,
        device: str = None,
        normalize_embeddings: bool = True,
        use_fp16: bool = False,
        return_dense: bool = True,
        return_sparse: bool = True,
        return_colbert_vecs: bool = False,
        **kwargs,
    ):
        import_datasets()
        import_FlagEmbedding()

        try:
            from FlagEmbedding import BGEM3FlagModel
        except AttributeError as e:
            import sys
            if "google.colab" in sys.modules and "ListView" in str(e):
                print(
                    "\033[91mIt looks like you're running on Google Colab. Please restart the session to resolve this issue.\033[0m")
                print(
                    "\033[91mFor further details, visit: https://github.com/milvus-io/milvus-model/issues/32.\033[0m")
            raise

        self.model_name = model_name
        self.batch_size = batch_size
        self.normalize_embeddings = normalize_embeddings
        self.device = device
        self.use_fp16 = use_fp16

        if device == "cpu" and use_fp16 is True:
            logger.warning(
                "Using fp16 with CPU can lead to runtime errors such as 'LayerNormKernelImpl', It's recommended to set 'use_fp16 = False' when using cpu. "
            )

        if "devices" in kwargs:
            device = kwargs["devices"]
            kwargs.pop("devices")

        _model_config = dict(
            {
                "model_name_or_path": model_name,
                "devices": device,
                "normalize_embeddings": normalize_embeddings,
                "use_fp16": use_fp16,
            },
            **kwargs,
        )
        _encode_config = {
            "batch_size": batch_size,
            "return_dense": return_dense,
            "return_sparse": return_sparse,
            "return_colbert_vecs": return_colbert_vecs,
        }
        self._model_config = _model_config
        self._encode_config = _encode_config

        self.model = BGEM3FlagModel(**self._model_config)

    def __call__(self, texts: List[str]) -> Dict:
        return self._encode(texts)

    @property
    def dim(self) -> Dict:
        return {
            "dense": self.model.model.model.config.hidden_size,
            "colbert_vecs": self.model.model.colbert_linear.out_features,
            "sparse": len(self.model.tokenizer),
        }

    def _encode(self, texts: List[str]) -> Dict:
        output = self.model.encode(sentences=texts, **self._encode_config)
        results = {}
        if self._encode_config["return_dense"] is True:
            results["dense"] = list(output["dense_vecs"])
        if self._encode_config["return_sparse"] is True:
            sparse_dim = self.dim["sparse"]
            results["sparse"] = []
            for sparse_vec in output["lexical_weights"]:
                indices = [int(k) for k in sparse_vec]
                values = np.array(list(sparse_vec.values()), dtype=np.float64)
                row_indices = [0] * len(indices)
                csr = csr_array((values, (row_indices, indices)), shape=(1, sparse_dim))
                results["sparse"].append(csr)
            results["sparse"] = stack_sparse_embeddings(results["sparse"]).tocsr()
        if self._encode_config["return_colbert_vecs"] is True:
            results["colbert_vecs"] = output["colbert_vecs"]
        return results


    def encode_queries(self, queries: List[str]) -> Dict:
        return self._encode(queries)

    def encode_documents(self, documents: List[str]) -> Dict:
        return self._encode(documents)




import onnxruntime

from transformers import AutoTokenizer, AutoConfig
import numpy as np
from typing import List

from pymilvus.model.base import BaseEmbeddingFunction
from pymilvus.model.utils import import_huggingface_hub


class OnnxEmbeddingFunction(BaseEmbeddingFunction):
    def __init__(self, model_name: str = "GPTCache/paraphrase-albert-onnx", tokenizer_name: str = "GPTCache/paraphrase-albert-small-v2"):
        import_huggingface_hub()
        from huggingface_hub import hf_hub_download

        self.tokenizer = AutoTokenizer.from_pretrained(tokenizer_name)
        self.model_name = model_name
        onnx_model_path = hf_hub_download(repo_id=model_name, filename="model.onnx")
        self.ort_session = onnxruntime.InferenceSession(onnx_model_path)
        config = AutoConfig.from_pretrained(
           tokenizer_name
        )
        self.__dimension = config.hidden_size

    def __call__(self, texts: List[str]) -> List[np.array]:
        return self._encode(texts)

    def encode_queries(self, queries: List[str]) -> List[np.array]:
        return self._encode(queries)

    def encode_documents(self, documents: List[str]) -> List[np.array]:
        return self._encode(documents)

    def _encode(self, texts: List[str]) -> List[np.array]:
        return [self._to_embedding(text) for text in texts]

    def _to_embedding(self, data: str, **_):
        encoded_text = self.tokenizer.encode_plus(data, padding="max_length", truncation=True)

        ort_inputs = {
            "input_ids": np.array(encoded_text["input_ids"]).astype("int64").reshape(1, -1),
            "attention_mask": np.array(encoded_text["attention_mask"]).astype("int64").reshape(1, -1),
            "token_type_ids": np.array(encoded_text["token_type_ids"]).astype("int64").reshape(1, -1),
        }

        ort_outputs = self.ort_session.run(None, ort_inputs)
        ort_feat = ort_outputs[0]
        emb = self._post_proc(ort_feat, ort_inputs["attention_mask"])
        emb =  emb.flatten()
        return emb / np.linalg.norm(emb)

    def _post_proc(self, token_embeddings, attention_mask):
        input_mask_expanded = (
            np.expand_dims(attention_mask, -1)
            .repeat(token_embeddings.shape[-1], -1)
            .astype(float)
        )
        sentence_embs = np.sum(token_embeddings * input_mask_expanded, 1) / np.maximum(
            input_mask_expanded.sum(1), 1e-9
        )
        return sentence_embs

    @property
    def dim(self):
        return self.__dimension




import os
from typing import List, Optional

import numpy as np
import requests

from pymilvus.model.base import BaseEmbeddingFunction

API_URL = "https://api.jina.ai/v1/embeddings"


class JinaEmbeddingFunction(BaseEmbeddingFunction):
    def __init__(
        self,
        model_name: str = "jina-embeddings-v3",
        api_key: Optional[str] = None,
        task: str = 'retrieval.passage',
        dimensions: Optional[int] = None,
        late_chunking: Optional[bool] = False,
        **kwargs,
    ):
        if api_key is None:
            if "JINAAI_API_KEY" in os.environ and os.environ["JINAAI_API_KEY"]:
                self.api_key = os.environ["JINAAI_API_KEY"]
            else:
                error_message = (
                    "Did not find api_key, please add an environment variable"
                    " `JINAAI_API_KEY` which contains it, or pass"
                    "  `api_key` as a named parameter."
                )
                raise ValueError(error_message)
        else:
            self.api_key = api_key
        self.model_name = model_name
        self._session = requests.Session()
        self._session.headers.update(
            {"Authorization": f"Bearer {self.api_key}", "Accept-Encoding": "identity"}
        )
        self.task = task
        self._dim = dimensions
        self.late_chunking = late_chunking

    @property
    def dim(self):
        if self._dim is None:
            self._dim = self._call_jina_api([""])[0].shape[0]
        return self._dim

    def encode_queries(self, queries: List[str]) -> List[np.array]:
        return self._call_jina_api(queries, task='retrieval.query')

    def encode_documents(self, documents: List[str]) -> List[np.array]:
        return self._call_jina_api(documents, task='retrieval.passage')

    def __call__(self, texts: List[str]) -> List[np.array]:
        return self._call_jina_api(texts, task=self.task)

    def _call_jina_api(self, texts: List[str], task: Optional[str] = None):
        data = {
            "input": texts,
            "model": self.model_name,
            "task": task,
            "late_chunking": self.late_chunking,
        }
        if self._dim is not None:
            data["dimensions"] = self._dim
        resp = self._session.post(  # type: ignore[assignment]
            API_URL,
            json=data,
        ).json()
        if "data" not in resp:
            raise RuntimeError(resp["detail"])

        embeddings = resp["data"]

        # Sort resulting embeddings by index
        sorted_embeddings = sorted(embeddings, key=lambda e: e["index"])  # type: ignore[no-any-return]
        return [np.array(result["embedding"]) for result in sorted_embeddings]



"""
The following code is adapted from/inspired by the 'neural-cherche' project:
https://github.com/raphaelsty/neural-cherche
Specifically, neural-cherche/neural_cherche/models/splade.py

MIT License

Copyright (c) 2023 Raphael Sourty

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
"""

from typing import Dict, List, Optional

from pymilvus.model.utils import import_transformers, import_scipy, import_torch
from pymilvus.model.sparse.utils import stack_sparse_embeddings

import_torch()
import_scipy()
import_transformers()

import torch
from scipy.sparse import csr_array
from transformers import AutoModelForMaskedLM, AutoTokenizer


class _SpladeImplementation:
    def __init__(
        self,
        model_name_or_path: Optional[str] = None,
        device: Optional[str] = None,
        batch_size: int = 32,
        **kwargs,
    ):
        self.device = device
        self.tokenizer = AutoTokenizer.from_pretrained(model_name_or_path)
        self.model = AutoModelForMaskedLM.from_pretrained(model_name_or_path, **kwargs)
        self.model.to(self.device)
        self.batch_size = batch_size

        self.relu = torch.nn.ReLU()
        self.relu.to(self.device)
        self.model.config.output_hidden_states = True

    def _encode(self, texts: List[str]):
        encoded_input = self.tokenizer.batch_encode_plus(
            texts,
            truncation=True,
            max_length=self.tokenizer.model_max_length,
            return_tensors="pt",
            add_special_tokens=True,
            padding=True,
        )
        encoded_input = {key: val.to(self.device) for key, val in encoded_input.items()}
        output = self.model(**encoded_input)
        return output.logits

    def _batchify(self, texts: List[str], batch_size: int) -> List[List[str]]:
        return [texts[i : i + batch_size] for i in range(0, len(texts), batch_size)]

    def forward(self, texts: List[str], k_tokens: int) -> csr_array:
        with torch.no_grad():
            batched_texts = self._batchify(texts, self.batch_size)
            sparse_embs = []
            for batch_texts in batched_texts:
                logits = self._encode(texts=batch_texts)
                activations = self._get_activation(logits=logits)
                if k_tokens is None:
                    nonzero_indices = torch.nonzero(activations["sparse_activations"])
                    activations["activations"] = nonzero_indices
                else:
                    activations = self._update_activations(**activations, k_tokens=k_tokens)
                batch_csr = self._convert_to_csr_array(activations)
                sparse_embs.extend(batch_csr)
        return stack_sparse_embeddings(sparse_embs).tocsr()

    def _get_activation(self, logits: torch.Tensor) -> Dict[str, torch.Tensor]:
        return {"sparse_activations": torch.amax(torch.log1p(self.relu(logits)), dim=1)}

    def _update_activations(self, sparse_activations: torch.Tensor, k_tokens: int) -> torch.Tensor:
        activations = torch.topk(input=sparse_activations, k=k_tokens, dim=1).indices

        # Set value of max sparse_activations which are not in top k to 0.
        sparse_activations = sparse_activations * torch.zeros(
            (sparse_activations.shape[0], sparse_activations.shape[1]),
            dtype=int,
            device=self.device,
        ).scatter_(dim=1, index=activations.long(), value=1)

        activations = torch.cat(
            (
                torch.arange(activations.shape[0], device=activations.device)
                .repeat_interleave(activations.shape[1])
                .reshape(-1, 1),
                activations.reshape((-1, 1)),
            ),
            dim=1,
        )

        return {
            "activations": activations,
            "sparse_activations": sparse_activations,
        }

    def _filter_activations(
        self, activations: torch.Tensor, k_tokens: int, **kwargs
    ) -> torch.Tensor:
        _, activations = torch.topk(input=activations, k=k_tokens, dim=1, **kwargs)
        return activations

    def _convert_to_csr_array(self, activations: Dict):

        values = (
            activations["sparse_activations"][
                activations["activations"][:, 0], activations["activations"][:, 1]
            ]
            .cpu()
            .detach()
            .numpy()
        )

        row_indices = activations["activations"][:, 0].cpu().detach().numpy()
        col_indices = activations["activations"][:, 1].cpu().detach().numpy()
        return csr_array(
            (values.flatten(), (row_indices, col_indices)),
            shape=activations["sparse_activations"].shape,
        )



import os
from typing import List, Optional

import requests

from pymilvus.model.base import BaseRerankFunction, RerankResult

API_URL = "https://api.jina.ai/v1/rerank"


class JinaRerankFunction(BaseRerankFunction):
    def __init__(self, model_name: str = "jina-reranker-v2-base-multilingual", api_key: Optional[str] = None):
        if api_key is None:
            if "JINAAI_API_KEY" in os.environ and os.environ["JINAAI_API_KEY"]:
                self.api_key = os.environ["JINAAI_API_KEY"]
            else:
                error_message = (
                    "Did not find api_key, please add an environment variable"
                    " `JINAAI_API_KEY` which contains it, or pass"
                    "  `api_key` as a named parameter."
                )
                raise ValueError(error_message)
        else:
            self.api_key = api_key
        self.model_name = model_name
        self._session = requests.Session()
        self._session.headers.update(
            {"Authorization": f"Bearer {self.api_key}", "Accept-Encoding": "identity"}
        )
        self.model_name = model_name

    def __call__(self, query: str, documents: List[str], top_k: int = 5) -> List[RerankResult]:
        resp = self._session.post(  # type: ignore[assignment]
            API_URL,
            json={
                "query": query,
                "documents": documents,
                "model": self.model_name,
                "top_n": top_k,
            },
        ).json()
        if "results" not in resp:
            raise RuntimeError(resp["detail"])

        results = []
        for res in resp["results"]:
            results.append(
                RerankResult(
                    text=res["document"]["text"], score=res["relevance_score"], index=res["index"]
                )
            )
        return results



from typing import List, Optional

from pymilvus.model.base import BaseRerankFunction, RerankResult
from pymilvus.model.utils import import_cohere


class CohereRerankFunction(BaseRerankFunction):
    def __init__(self, model_name: str = "rerank-english-v3.0", api_key: Optional[str] = None, return_documents=True, **kwargs):
        import_cohere()
        import cohere

        self.model_name = model_name
        self.client = cohere.ClientV2(api_key)
        self.rerank_config = {"return_documents": return_documents, **kwargs} 


    def __call__(self, query: str, documents: List[str], top_k: int = 5) -> List[RerankResult]:
        co_results = self.client.rerank(
            query=query, documents=documents, top_n=top_k, model=self.model_name, **self.rerank_config)
        results = []
        for co_result in co_results.results:
            document_text = ""
            if self.rerank_config["return_documents"] is True:
                document_text = co_result.document.text
            results.append(
                RerankResult(
                    text=document_text,
                    score=co_result.relevance_score,
                    index=co_result.index,
                )
            )
        return results



from pymilvus.model.dense.openai import OpenAIEmbeddingFunction
from pymilvus.model.dense.sentence_transformer import SentenceTransformerEmbeddingFunction
from pymilvus.model.dense.voyageai import VoyageEmbeddingFunction
from pymilvus.model.dense.jinaai import JinaEmbeddingFunction
from pymilvus.model.dense.tei import TEIEmbeddingFunction
from pymilvus.model.dense.onnx import OnnxEmbeddingFunction
from pymilvus.model.dense.cohere import CohereEmbeddingFunction
from pymilvus.model.dense.mistralai import MistralAIEmbeddingFunction
from pymilvus.model.dense.nomic import NomicEmbeddingFunction
from pymilvus.model.dense.instructor import InstructorEmbeddingFunction
from pymilvus.model.dense.model2vec_embed import Model2VecEmbeddingFunction

__all__ = [
    "OpenAIEmbeddingFunction",
    "SentenceTransformerEmbeddingFunction",
    "VoyageEmbeddingFunction",
    "JinaEmbeddingFunction",
    "TEIEmbeddingFunction",
    "OnnxEmbeddingFunction",
    "CohereEmbeddingFunction",
    "MistralAIEmbeddingFunction",
    "NomicEmbeddingFunction",
    "InstructorEmbeddingFunction",
    "Model2VecEmbeddingFunction",
]



#The following code is adapted from/inspired by the 'instructor-embedding' project:
#https://github.com/xlang-ai/instructor-embedding
#Specifically, instructor-embedding/InstructorEmbedding/instructor.py
#
# License: Apache License 2.0 (January 2004)
#
# For more information on the original project, visit the GitHub repository:
# https://github.com/xlang-ai/instructor-embedding
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at:
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# This script is based on the modifications from https://github.com/UKPLab/sentence-transformers
import importlib
import json
import os
from collections import OrderedDict
from typing import Union

from pymilvus.model.utils import import_sentence_transformers, import_huggingface_hub, import_torch

import_sentence_transformers()
import_huggingface_hub()
import_torch()

import numpy as np
import torch
from sentence_transformers import SentenceTransformer
from sentence_transformers.models import Transformer
from torch import Tensor, nn
from transformers import AutoConfig, AutoTokenizer
from sentence_transformers.util import disabled_tqdm
from huggingface_hub import snapshot_download


def batch_to_device(batch, target_device: str):
    for key in batch:
        if isinstance(batch[key], Tensor):
            batch[key] = batch[key].to(target_device)
    return batch


class InstructorPooling(nn.Module):
    """Performs pooling (max or mean) on the token embeddings.

    Using pooling, it generates from a variable sized sentence a fixed sized sentence embedding.
    This layer also allows to use the CLS token if it is returned by the underlying word embedding model.
    You can concatenate multiple poolings together.

    :param word_embedding_dimension: Dimensions for the word embeddings
    :param pooling_mode: Can be a string: mean/max/cls. If set, overwrites the other pooling_mode_* settings
    :param pooling_mode_cls_token: Use the first token (CLS token) as text representations
    :param pooling_mode_max_tokens: Use max in each dimension over all tokens.
    :param pooling_mode_mean_tokens: Perform mean-pooling
    :param pooling_mode_mean_sqrt_len_tokens: Perform mean-pooling, but divide by sqrt(input_length).
    :param pooling_mode_weightedmean_tokens: Perform (position) weighted mean pooling,
    see https://arxiv.org/abs/2202.08904
    :param pooling_mode_lasttoken: Perform last token pooling,
    see https://arxiv.org/abs/2202.08904 & https://arxiv.org/abs/2201.10005
    """

    def __init__(
        self,
        word_embedding_dimension: int,
        pooling_mode: Union[str, None] = None,
        pooling_mode_cls_token: bool = False,
        pooling_mode_max_tokens: bool = False,
        pooling_mode_mean_tokens: bool = True,
        pooling_mode_mean_sqrt_len_tokens: bool = False,
        pooling_mode_weightedmean_tokens: bool = False,
        pooling_mode_lasttoken: bool = False,
    ):
        super().__init__()

        self.config_keys = [
            "word_embedding_dimension",
            "pooling_mode_cls_token",
            "pooling_mode_mean_tokens",
            "pooling_mode_max_tokens",
            "pooling_mode_mean_sqrt_len_tokens",
            "pooling_mode_weightedmean_tokens",
            "pooling_mode_lasttoken",
        ]

        if pooling_mode is not None:  # Set pooling mode by string
            pooling_mode = pooling_mode.lower()
            assert pooling_mode in ["mean", "max", "cls", "weightedmean", "lasttoken"]
            pooling_mode_cls_token = pooling_mode == "cls"
            pooling_mode_max_tokens = pooling_mode == "max"
            pooling_mode_mean_tokens = pooling_mode == "mean"
            pooling_mode_weightedmean_tokens = pooling_mode == "weightedmean"
            pooling_mode_lasttoken = pooling_mode == "lasttoken"

        self.word_embedding_dimension = word_embedding_dimension
        self.pooling_mode_cls_token = pooling_mode_cls_token
        self.pooling_mode_mean_tokens = pooling_mode_mean_tokens
        self.pooling_mode_max_tokens = pooling_mode_max_tokens
        self.pooling_mode_mean_sqrt_len_tokens = pooling_mode_mean_sqrt_len_tokens
        self.pooling_mode_weightedmean_tokens = pooling_mode_weightedmean_tokens
        self.pooling_mode_lasttoken = pooling_mode_lasttoken

        pooling_mode_multiplier = sum(
            [
                pooling_mode_cls_token,
                pooling_mode_max_tokens,
                pooling_mode_mean_tokens,
                pooling_mode_mean_sqrt_len_tokens,
                pooling_mode_weightedmean_tokens,
                pooling_mode_lasttoken,
            ]
        )
        self.pooling_output_dimension = (
            pooling_mode_multiplier * word_embedding_dimension
        )

    def __repr__(self):
        return f"Pooling({self.get_config_dict()})"

    def get_pooling_mode_str(self) -> str:
        """
        Returns the pooling mode as string
        """
        modes = []
        if self.pooling_mode_cls_token:
            modes.append("cls")
        if self.pooling_mode_mean_tokens:
            modes.append("mean")
        if self.pooling_mode_max_tokens:
            modes.append("max")
        if self.pooling_mode_mean_sqrt_len_tokens:
            modes.append("mean_sqrt_len_tokens")
        if self.pooling_mode_weightedmean_tokens:
            modes.append("weightedmean")
        if self.pooling_mode_lasttoken:
            modes.append("lasttoken")

        return "+".join(modes)

    def forward(self, features):
        # print(features.keys())
        token_embeddings = features["token_embeddings"]
        attention_mask = features["attention_mask"]

        ## Pooling strategy
        output_vectors = []
        if self.pooling_mode_cls_token:
            cls_token = features.get(
                "cls_token_embeddings", token_embeddings[:, 0]
            )  # Take first token by default
            output_vectors.append(cls_token)
        if self.pooling_mode_max_tokens:
            input_mask_expanded = (
                attention_mask.unsqueeze(-1).expand(token_embeddings.size()).float()
            )
            token_embeddings[
                input_mask_expanded == 0
            ] = -1e9  # Set padding tokens to large negative value
            max_over_time = torch.max(token_embeddings, 1)[0]
            output_vectors.append(max_over_time)
        if self.pooling_mode_mean_tokens or self.pooling_mode_mean_sqrt_len_tokens:
            input_mask_expanded = (
                attention_mask.unsqueeze(-1).expand(token_embeddings.size()).float()
            )
            sum_embeddings = torch.sum(token_embeddings * input_mask_expanded, 1)

            # If tokens are weighted (by WordWeights layer), feature 'token_weights_sum' will be present
            if "token_weights_sum" in features:
                sum_mask = (
                    features["token_weights_sum"]
                    .unsqueeze(-1)
                    .expand(sum_embeddings.size())
                )
            else:
                sum_mask = input_mask_expanded.sum(1)

            sum_mask = torch.clamp(sum_mask, min=1e-9)

            if self.pooling_mode_mean_tokens:
                output_vectors.append(sum_embeddings / sum_mask)
            if self.pooling_mode_mean_sqrt_len_tokens:
                output_vectors.append(sum_embeddings / torch.sqrt(sum_mask))
        if self.pooling_mode_weightedmean_tokens:
            input_mask_expanded = (
                attention_mask.unsqueeze(-1).expand(token_embeddings.size()).float()
            )
            # token_embeddings shape: bs, seq, hidden_dim
            weights = (
                torch.arange(start=1, end=token_embeddings.shape[1] + 1)
                .unsqueeze(0)
                .unsqueeze(-1)
                .expand(token_embeddings.size())
                .float()
                .to(token_embeddings.device)
            )
            assert weights.shape == token_embeddings.shape == input_mask_expanded.shape
            input_mask_expanded = input_mask_expanded * weights

            sum_embeddings = torch.sum(token_embeddings * input_mask_expanded, 1)

            # If tokens are weighted (by WordWeights layer), feature 'token_weights_sum' will be present
            if "token_weights_sum" in features:
                sum_mask = (
                    features["token_weights_sum"]
                    .unsqueeze(-1)
                    .expand(sum_embeddings.size())
                )
            else:
                sum_mask = input_mask_expanded.sum(1)

            sum_mask = torch.clamp(sum_mask, min=1e-9)
            output_vectors.append(sum_embeddings / sum_mask)
        if self.pooling_mode_lasttoken:
            batch_size, _, hidden_dim = token_embeddings.shape
            # attention_mask shape: (bs, seq_len)
            # Get shape [bs] indices of the last token (i.e. the last token for each batch item)
            # argmin gives us the index of the first 0 in the attention mask;
            # We get the last 1 index by subtracting 1
            gather_indices = (
                torch.argmin(attention_mask, 1, keepdim=False) - 1
            )  # Shape [bs]

            # There are empty sequences, where the index would become -1 which will crash
            gather_indices = torch.clamp(gather_indices, min=0)

            # Turn indices from shape [bs] --> [bs, 1, hidden_dim]
            gather_indices = gather_indices.unsqueeze(-1).repeat(1, hidden_dim)
            gather_indices = gather_indices.unsqueeze(1)
            assert gather_indices.shape == (batch_size, 1, hidden_dim)

            # Gather along the 1st dim (seq_len) (bs, seq_len, hidden_dim -> bs, hidden_dim)
            # Actually no need for the attention mask as we gather the last token where attn_mask = 1
            # but as we set some indices (which shouldn't be attended to) to 0 with clamp, we
            # use the attention mask to ignore them again
            input_mask_expanded = (
                attention_mask.unsqueeze(-1).expand(token_embeddings.size()).float()
            )
            embedding = torch.gather(
                token_embeddings * input_mask_expanded, 1, gather_indices
            ).squeeze(dim=1)
            output_vectors.append(embedding)

        output_vector = torch.cat(output_vectors, 1)
        features.update({"sentence_embedding": output_vector})
        return features

    def get_sentence_embedding_dimension(self):
        return self.pooling_output_dimension

    def get_config_dict(self):
        return {key: self.__dict__[key] for key in self.config_keys}

    def save(self, output_path):
        with open(
            os.path.join(output_path, "config.json"), "w", encoding="UTF-8"
        ) as config_file:
            json.dump(self.get_config_dict(), config_file, indent=2)

    @staticmethod
    def load(input_path):
        with open(
            os.path.join(input_path, "config.json"), encoding="UTF-8"
        ) as config_file:
            config = json.load(config_file)

        return InstructorPooling(**config)


def import_from_string(dotted_path):
    """
    Import a dotted module path and return the attribute/class designated by the
    last name in the path. Raise ImportError if the import failed.
    """
    try:
        module_path, class_name = dotted_path.rsplit(".", 1)
    except ValueError:
        msg = f"{dotted_path} doesn't look like a module path"
        raise ImportError(msg)

    try:
        module = importlib.import_module(dotted_path)
    except:
        module = importlib.import_module(module_path)

    try:
        return getattr(module, class_name)
    except AttributeError:
        msg = f"Module {module_path} does not define a {class_name} attribute/class"
        raise ImportError(msg)


class InstructorTransformer(Transformer):
    def __init__(
        self,
        model_name_or_path: str,
        max_seq_length=None,
        model_args=None,
        cache_dir=None,
        tokenizer_args=None,
        do_lower_case: bool = False,
        tokenizer_name_or_path: Union[str, None] = None,
        load_model: bool = True,
    ):
        super().__init__(model_name_or_path)
        if model_args is None:
            model_args = {}
        if tokenizer_args is None:
            tokenizer_args = {}
        self.config_keys = ["max_seq_length", "do_lower_case"]
        self.do_lower_case = do_lower_case
        self.model_name_or_path = model_name_or_path
        if model_name_or_path == "bi-contriever":
            model_name_or_path = "facebook/contriever"
        if model_name_or_path.startswith("bigtr"):
            model_name_or_path = model_name_or_path.split("#")[1]
        if "bigtr" in model_name_or_path and os.path.isdir(model_name_or_path):
            config = AutoConfig.from_pretrained(
                os.path.join(model_name_or_path, "with_prompt"),
                **model_args,
                cache_dir=cache_dir,
            )
        else:
            config = AutoConfig.from_pretrained(
                model_name_or_path, **model_args, cache_dir=cache_dir
            )

        if load_model:
            self._load_model(self.model_name_or_path, config, cache_dir, **model_args)
        self.tokenizer = AutoTokenizer.from_pretrained(
            tokenizer_name_or_path
            if tokenizer_name_or_path is not None
            else model_name_or_path,
            cache_dir=cache_dir,
            **tokenizer_args,
        )

        if max_seq_length is None:
            if (
                hasattr(self.auto_model, "config")
                and hasattr(self.auto_model.config, "max_position_embeddings")
                and hasattr(self.tokenizer, "model_max_length")
            ):
                max_seq_length = min(
                    self.auto_model.config.max_position_embeddings,
                    self.tokenizer.model_max_length,
                )

        self.max_seq_length = max_seq_length
        if tokenizer_name_or_path is not None:
            self.auto_model.config.tokenizer_class = self.tokenizer.__class__.__name__

    def forward(self, features):
        input_features = {
            "input_ids": features["input_ids"],
            "attention_mask": features["attention_mask"],
        }
        if "token_type_ids" in features:
            input_features["token_type_ids"] = features["token_type_ids"]

        instruction_mask = features["instruction_mask"]
        output_states = self.auto_model(**input_features, return_dict=False)
        output_tokens = output_states[0]
        attention_mask = features["attention_mask"]
        instruction_mask = features["instruction_mask"]
        attention_mask = attention_mask * instruction_mask
        features.update(
            {"token_embeddings": output_tokens, "attention_mask": attention_mask}
        )

        if self.auto_model.config.output_hidden_states:
            all_layer_idx = 2
            if (
                len(output_states) < 3
            ):  # Some models only output last_hidden_states and all_hidden_states
                all_layer_idx = 1
            hidden_states = output_states[all_layer_idx]
            features.update({"all_layer_embeddings": hidden_states})

        return features

    @staticmethod
    def load(input_path: str):
        # Old classes used other config names than 'sentence_bert_config.json'
        for config_name in [
            "sentence_bert_config.json",
            "sentence_roberta_config.json",
            "sentence_distilbert_config.json",
            "sentence_camembert_config.json",
            "sentence_albert_config.json",
            "sentence_xlm-roberta_config.json",
            "sentence_xlnet_config.json",
        ]:
            sbert_config_path = os.path.join(input_path, config_name)
            if os.path.exists(sbert_config_path):
                break

        with open(sbert_config_path, encoding="UTF-8") as config_file:
            config = json.load(config_file)
        return InstructorTransformer(model_name_or_path=input_path, **config)

    def tokenize(self, texts):
        """
        Tokenizes a text and maps tokens to token-ids
        """
        output = {}
        if isinstance(texts[0], str):
            to_tokenize = [texts]
            to_tokenize = [[str(s).strip() for s in col] for col in to_tokenize]

            # Lowercase
            if self.do_lower_case:
                to_tokenize = [[s.lower() for s in col] for col in to_tokenize]

            input_features = self.tokenizer(
                *to_tokenize,
                padding="max_length",
                truncation="longest_first",
                return_tensors="pt",
                max_length=self.max_seq_length,
            )

        elif isinstance(texts[0], list):
            assert isinstance(texts[0][1], str)
            assert (
                len(texts[0]) == 2
            ), "The input should have both instruction and input text"

            instructions = []
            instruction_prepended_input_texts = []
            for pair in texts:
                instruction = pair[0].strip()
                text = pair[1].strip()
                if self.do_lower_case:
                    instruction = instruction.lower()
                    text = text.lower()
                instructions.append(instruction)
                instruction_prepended_input_texts.append("".join([instruction, text]))

            input_features = self.tokenize(instruction_prepended_input_texts)
            instruction_features = self.tokenize(instructions)
            input_features = Instructor.prepare_input_features(
                input_features, instruction_features
            )
        else:
            raise ValueError("not support other modes")

        output.update(input_features)
        return output


class Instructor(SentenceTransformer):
    @staticmethod
    def prepare_input_features(
        input_features, instruction_features, return_data_type: str = "pt"
    ):
        if return_data_type == "np":
            input_features["attention_mask"] = torch.from_numpy(
                input_features["attention_mask"]
            )
            instruction_features["attention_mask"] = torch.from_numpy(
                instruction_features["attention_mask"]
            )

        input_attention_mask_shape = input_features["attention_mask"].shape
        instruction_attention_mask = instruction_features["attention_mask"]

        # reducing the attention length by 1 in order to omit the attention corresponding to the end_token
        instruction_attention_mask = instruction_attention_mask[:, 1:]

        # creating instruction attention matrix equivalent to the size of the input attention matrix
        expanded_instruction_attention_mask = torch.zeros(
            input_attention_mask_shape, dtype=torch.int64
        )
        # assigning the the actual instruction attention matrix to the expanded_instruction_attention_mask
        # eg:
        # instruction_attention_mask: 3x3
        #  [[1,1,1],
        #   [1,1,0],
        #   [1,0,0]]
        # expanded_instruction_attention_mask: 3x4
        #  [[1,1,1,0],
        #   [1,1,0,0],
        #   [1,0,0,0]]
        expanded_instruction_attention_mask[
            : instruction_attention_mask.size(0), : instruction_attention_mask.size(1)
        ] = instruction_attention_mask

        # In the pooling layer we want to consider only the tokens corresponding to the input text
        # and not the instruction. This is achieved by inverting the
        # attention_mask corresponding to the instruction.
        expanded_instruction_attention_mask = 1 - expanded_instruction_attention_mask
        input_features["instruction_mask"] = expanded_instruction_attention_mask
        if return_data_type == "np":
            input_features["attention_mask"] = input_features["attention_mask"].numpy()
            instruction_features["attention_mask"] = instruction_features[
                "attention_mask"
            ].numpy()
        return input_features

    def smart_batching_collate(self, batch):
        num_texts = len(batch[0].texts)
        texts = [[] for _ in range(num_texts)]
        labels = []

        for example in batch:
            for idx, text in enumerate(example.texts):
                texts[idx].append(text)
            labels.append(example.label)

        labels = torch.tensor(labels)
        batched_input_features = []

        for idx in range(num_texts):
            assert isinstance(texts[idx][0], list)
            assert (
                len(texts[idx][0]) == 2
            ), "The input should have both instruction and input text"

            num = len(texts[idx])
            instructions = []
            instruction_prepended_input_texts = []
            for local_idx in range(num):
                assert len(texts[idx][local_idx]) == 2
                instructions.append(texts[idx][local_idx][0])
                instruction_prepended_input_texts.append("".join(texts[idx][local_idx]))
                assert isinstance(instructions[-1], str)
                assert isinstance(instruction_prepended_input_texts[-1], str)

            input_features = self.tokenize(instruction_prepended_input_texts)
            instruction_features = self.tokenize(instructions)
            input_features = Instructor.prepare_input_features(
                input_features, instruction_features
            )
            batched_input_features.append(input_features)

        return batched_input_features, labels

    def _load_sbert_model(self, model_path, token = None, cache_folder = None, revision = None, trust_remote_code = False, **kwargs):
        """
        Loads a full sentence-transformers model
        """
        # Taken mostly from: https://github.com/UKPLab/sentence-transformers/blob/66e0ee30843dd411c64f37f65447bb38c7bf857a/sentence_transformers/util.py#L544
        download_kwargs = {
            "repo_id": model_path,
            "revision": revision,
            "library_name": "sentence-transformers",
            "token": token,
            "cache_dir": cache_folder,
            "tqdm_class": disabled_tqdm,
        }
        model_path = snapshot_download(**download_kwargs)

        # Check if the config_sentence_transformers.json file exists (exists since v2 of the framework)
        config_sentence_transformers_json_path = os.path.join(
            model_path, "config_sentence_transformers.json"
        )
        if os.path.exists(config_sentence_transformers_json_path):
            with open(
                config_sentence_transformers_json_path, encoding="UTF-8"
            ) as config_file:
                self._model_config = json.load(config_file)

        # Check if a readme exists
        model_card_path = os.path.join(model_path, "README.md")
        if os.path.exists(model_card_path):
            try:
                with open(model_card_path, encoding="utf8") as config_file:
                    self._model_card_text = config_file.read()
            except:
                pass

        # Load the modules of sentence transformer
        modules_json_path = os.path.join(model_path, "modules.json")
        with open(modules_json_path, encoding="UTF-8") as config_file:
            modules_config = json.load(config_file)

        modules = OrderedDict()
        for module_config in modules_config:
            if module_config["idx"] == 0:
                module_class = InstructorTransformer
            elif module_config["idx"] == 1:
                module_class = InstructorPooling
            else:
                module_class = import_from_string(module_config["type"])
            module = module_class.load(os.path.join(model_path, module_config["path"]))
            modules[module_config["name"]] = module

        return modules

    def encode(
        self,
        sentences,
        batch_size: int = 32,
        show_progress_bar: Union[bool, None] = None,
        output_value: str = "sentence_embedding",
        convert_to_numpy: bool = True,
        convert_to_tensor: bool = False,
        device: Union[str, None] = None,
        normalize_embeddings: bool = False,
    ):
        """
        Computes sentence embeddings

        :param sentences: the sentences to embed
        :param batch_size: the batch size used for the computation
        :param show_progress_bar: Output a progress bar when encode sentences
        :param output_value:  Default sentence_embedding, to get sentence embeddings.
        Can be set to token_embeddings to get wordpiece token embeddings. Set to None, to get all output values
        :param convert_to_numpy: If true, the output is a list of numpy vectors.
        Else, it is a list of pytorch tensors.
        :param convert_to_tensor: If true, you get one large tensor as return.
        Overwrites any setting from convert_to_numpy
        :param device: Which torch.device to use for the computation
        :param normalize_embeddings: If set to true, returned vectors will have length 1.
        In that case, the faster dot-product (util.dot_score) instead of cosine similarity can be used.

        :return:
           By default, a list of tensors is returned. If convert_to_tensor,
           a stacked tensor is returned. If convert_to_numpy, a numpy matrix is returned.
        """
        self.eval()
        if show_progress_bar is None:
            show_progress_bar = False

        if convert_to_tensor:
            convert_to_numpy = False

        if output_value != "sentence_embedding":
            convert_to_tensor = False
            convert_to_numpy = False

        input_was_string = False
        if isinstance(sentences, str) or not hasattr(
            sentences, "__len__"
        ):  # Cast an individual sentence to a list with length 1
            sentences = [sentences]
            input_was_string = True

        if device is None:
            device = self.device

        self.to(device)

        all_embeddings = []
        if isinstance(sentences[0], list):
            lengths = []
            for sen in sentences:
                lengths.append(-self._text_length(sen[1]))
            length_sorted_idx = np.argsort(lengths)
        else:
            length_sorted_idx = np.argsort(
                [-self._text_length(sen) for sen in sentences]
            )
        sentences_sorted = [sentences[idx] for idx in length_sorted_idx]

        for start_index in range(0, len(sentences), batch_size):
            sentences_batch = sentences_sorted[start_index : start_index + batch_size]
            features = self.tokenize(sentences_batch)
            features = batch_to_device(features, device)

            with torch.no_grad():
                out_features = self.forward(features)

                if output_value == "token_embeddings":
                    embeddings = []
                    for token_emb, attention in zip(
                        out_features[output_value], out_features["attention_mask"]
                    ):
                        last_mask_id = len(attention) - 1
                        while last_mask_id > 0 and attention[last_mask_id].item() == 0:
                            last_mask_id -= 1

                        embeddings.append(token_emb[0 : last_mask_id + 1])
                elif output_value is None:  # Return all outputs
                    embeddings = []
                    for sent_idx in range(len(out_features["sentence_embedding"])):
                        row = {
                            name: out_features[name][sent_idx] for name in out_features
                        }
                        embeddings.append(row)
                else:  # Sentence embeddings
                    embeddings = out_features[output_value]
                    embeddings = embeddings.detach()
                    if normalize_embeddings:
                        embeddings = torch.nn.functional.normalize(
                            embeddings, p=2, dim=1
                        )

                    # fixes for #522 and #487 to avoid oom problems on gpu with large datasets
                    if convert_to_numpy:
                        embeddings = embeddings.cpu()

                all_embeddings.extend(embeddings)

        all_embeddings = [all_embeddings[idx] for idx in np.argsort(length_sorted_idx)]

        if convert_to_tensor:
            all_embeddings = torch.stack(all_embeddings)
        elif convert_to_numpy:
            all_embeddings = np.asarray([emb.numpy() for emb in all_embeddings])

        if input_was_string:
            all_embeddings = all_embeddings[0]

        return all_embeddings



from typing import List, Optional

from pymilvus.model.base import BaseRerankFunction, RerankResult
from pymilvus.model.utils import import_voyageai


class VoyageRerankFunction(BaseRerankFunction):
    def __init__(self, model_name: str = "rerank-2", api_key: Optional[str] = None):
        import_voyageai()
        import voyageai

        self.model_name = model_name
        self.client = voyageai.Client(api_key=api_key)

    def __call__(self, query: str, documents: List[str], top_k: int = 5) -> List[RerankResult]:
        vo_results = self.client.rerank(query, documents, model=self.model_name, top_k=top_k)
        results = []
        for vo_result in vo_results.results:
            results.append(
                RerankResult(
                    text=vo_result.document, score=vo_result.relevance_score, index=vo_result.index
                )
            )
        return results



__all__ = ["DefaultEmbeddingFunction", "dense", "sparse", "hybrid", "reranker", "utils"]

from . import dense, hybrid, sparse, reranker, utils

DefaultEmbeddingFunction = dense.onnx.OnnxEmbeddingFunction



from typing import Any, List

from pymilvus.model.base import BaseRerankFunction, RerankResult
from pymilvus.model.utils import import_sentence_transformers


class CrossEncoderRerankFunction(BaseRerankFunction):
    def __init__(
        self,
        model_name: str = "",
        device: str = "",
        batch_size: int = 32,
        activation_fct: Any = None,
        **kwargs,
    ):
        import_sentence_transformers()
        import sentence_transformers

        if sentence_transformers is None:
            error_message = "sentence_transformer is not installed."
            raise ImportError(error_message)
        self.model_name = model_name
        self.device = device
        self.batch_size = batch_size
        self.activation_fct = activation_fct
        self.model = sentence_transformers.cross_encoder.CrossEncoder(
            model_name=model_name, device=self.device, default_activation_function=activation_fct
        )

    def __call__(self, query: str, documents: List[str], top_k: int = 5) -> List[RerankResult]:
        query_document_pairs = [[query, doc] for doc in documents]
        scores = self.model.predict(query_document_pairs, batch_size=self.batch_size)

        ranked_order = sorted(range(len(scores)), key=lambda i: scores[i], reverse=True)
        if top_k:
            ranked_order = ranked_order[:top_k]

        results = []
        for index in ranked_order:
            results.append(RerankResult(text=documents[index], score=scores[index], index=index))
        return results



from typing import Any, List, Optional, Union

from pymilvus.model.base import BaseRerankFunction, RerankResult
from pymilvus.model.utils import import_FlagEmbedding, import_transformers


class BGERerankFunction(BaseRerankFunction):
    def __init__(
        self,
        model_name: str = "BAAI/bge-reranker-v2-m3",
        use_fp16: bool = True,
        batch_size: int = 32,
        normalize: bool = True,
        device: Optional[Union[str, List]] = None,
        query_max_length: int = 256,
        max_length: int = 512,
        **kwargs: Any,
    ):
        import_FlagEmbedding()
        import_transformers()
        from FlagEmbedding import FlagAutoReranker

        self.model_name = model_name
        self.batch_size = batch_size
        self.normalize = normalize
        self.device = device

        if "devices" in kwargs:
            device = kwargs["devices"]
            kwargs.pop("devices")

        _model_config = dict(
            {
                "model_name_or_path": model_name,
                "batch_size": batch_size,
                "use_fp16": use_fp16,
                "devices": device,
                "max_length": max_length,
                "query_max_length": query_max_length,
                "normalize": normalize,
            },
            **kwargs,
        )
        self.reranker = FlagAutoReranker.from_finetuned(**_model_config)


    def _batchify(self, texts: List[str], batch_size: int) -> List[List[str]]:
        return [texts[i : i + batch_size] for i in range(0, len(texts), batch_size)]

    def __call__(self, query: str, documents: List[str], top_k: int = 5) -> List[RerankResult]:
        query_document_pairs = [[query, doc] for doc in documents]
        batched_texts = self._batchify(documents, self.batch_size)
        scores = []
        for batched_text in batched_texts:
            query_document_pairs = [[query, text] for text in batched_text]
            batch_score = self.reranker.compute_score(
                query_document_pairs, normalize=self.normalize
            )
            scores.extend(batch_score)
        ranked_order = sorted(range(len(scores)), key=lambda i: scores[i], reverse=True)

        if top_k:
            ranked_order = ranked_order[:top_k]

        results = []
        for index in ranked_order:
            results.append(RerankResult(text=documents[index], score=scores[index], index=index))
        return results




"""
this module is a hack only in place to allow for setuptools
to use the attribute for the versions

it works only if the backend-path of the build-system section
from pyproject.toml is respected
"""

from __future__ import annotations

import logging
from typing import Callable

from setuptools import build_meta as build_meta  # noqa

from setuptools_scm import _types as _t
from setuptools_scm import Configuration
from setuptools_scm import get_version
from setuptools_scm import git
from setuptools_scm import hg
from setuptools_scm.fallbacks import parse_pkginfo
from setuptools_scm.version import (
    get_no_local_node,
    _parse_version_tag,
    guess_next_simple_semver,
    SEMVER_MINOR,
    guess_next_version,
    ScmVersion,
)

log = logging.getLogger("setuptools_scm")
# todo: take fake entrypoints from pyproject.toml
try_parse: list[Callable[[_t.PathT, Configuration], ScmVersion | None]] = [
    parse_pkginfo,
    git.parse,
    hg.parse,
    git.parse_archival,
    hg.parse_archival,
]


def parse(root: str, config: Configuration) -> ScmVersion | None:
    for maybe_parse in try_parse:
        try:
            parsed = maybe_parse(root, config)
        except OSError as e:
            log.warning("parse with %s failed with: %s", maybe_parse, e)
        else:
            if parsed is not None:
                return parsed


fmt = "{guessed}.rc{distance}"


def custom_version(version: ScmVersion) -> str:
    if version.exact:
        return version.format_with("{tag}")
    if version.branch is not None:
        # Does the branch name (stripped of namespace) parse as a version?
        branch_ver_data = _parse_version_tag(version.branch.split("/")[-1], version.config)
        if branch_ver_data is not None:
            branch_ver = branch_ver_data["version"]
            if branch_ver[0] == "v":
                # Allow branches that start with 'v', similar to Version.
                branch_ver = branch_ver[1:]
            # Does the branch version up to the minor part match the tag? If not it
            # might be like, an issue number or something and not a version number, so
            # we only want to use it if it matches.
            tag_ver_up_to_minor = str(version.tag).split(".")[:SEMVER_MINOR]
            branch_ver_up_to_minor = branch_ver.split(".")[:SEMVER_MINOR]
            if branch_ver_up_to_minor == tag_ver_up_to_minor:
                # We're in a release/maintenance branch, next is a patch/rc/beta bump:
                return version.format_next_version(guess_next_version, fmt=fmt)
    # We're in a development branch, next is a minor bump:
    # return version.format_next_version(guess_next_simple_semver, retain=SEMVER_MINOR, fmt=fmt)
    return version.format_next_version(guess_next_version, fmt=fmt)


def scm_version() -> str:
    return get_version(
        relative_to=__file__,
        parse=parse,
        version_scheme=custom_version,
        local_scheme=get_no_local_node,
    )


version: str


def __getattr__(name: str) -> str:
    if name == "version":
        global version
        version = scm_version()
        return version
    raise AttributeError(name)



"""
The following code is adapted from/inspired from :
https://huggingface.co/Alibaba-NLP/gte-multilingual-base/blob/main/scripts/gte_embedding.py

# Copyright 2024 The GTE Team Authors and Alibaba Group.
# Licensed under the Apache License, Version 2.0 (the "License");
"""
import logging

from typing import Dict, List, Optional

from pymilvus.model.base import BaseEmbeddingFunction
from pymilvus.model.sparse.utils import stack_sparse_embeddings

logger = logging.getLogger(__name__)
logger.setLevel(logging.DEBUG)



class MGTEEmbeddingFunction(BaseEmbeddingFunction):
    def __init__(
        self,
        model_name: str = "Alibaba-NLP/gte-multilingual-base",
        batch_size: int = 16,
        device: str = "",
        normalize_embeddings: bool = True,
        dimensions: Optional[int] = None,
        use_fp16: bool = False,
        return_dense: bool = True,
        return_sparse: bool = True,
        **kwargs,
    ):
        from .mgte_embedding.gte_impl import _GTEEmbeddidng
        self.model_name = model_name
        self.batch_size = batch_size
        self.normalize_embeddings = normalize_embeddings
        self.device = device
        self.use_fp16 = use_fp16
        self.dimensions = dimensions

        if "dimension" in kwargs: 
            self.dimensions = kwargs["dimension"]
            kwargs.pop("dimension")

        if device == "cpu" and use_fp16 is True:
            logger.warning(
                "Using fp16 with CPU can lead to runtime errors such as 'LayerNormKernelImpl', It's recommended to set 'use_fp16 = False' when using cpu. "
            )

        _model_config = dict(
            {
                "model_name": model_name,
                "device": device,
                "normalized": normalize_embeddings,
                "use_fp16": use_fp16,
            },
            **kwargs,
        )
        _encode_config = {
            "batch_size": batch_size,
            "return_dense": return_dense,
            "return_sparse": return_sparse,
        }
        self._model_config = _model_config
        self._encode_config = _encode_config

        self.model = _GTEEmbeddidng(**self._model_config)

        _encode_config["dimension"] = self.dimensions
        
        if self.dimensions is None:
            self.dimensions = self.model.model.config.hidden_size 

    def __call__(self, texts: List[str]) -> Dict:
        return self._encode(texts)

    @property
    def dim(self) -> Dict:
        return {
            "dense": self.dimensions,
            "sparse": len(self.model.tokenizer),
        }

    def _encode(self, texts: List[str]) -> Dict:
        from scipy.sparse import csr_array

        output = self.model.encode(texts=texts, **self._encode_config)
        results = {}
        if self._encode_config["return_dense"] is True:
            results["dense"] = list(output["dense_embeddings"])
        if self._encode_config["return_sparse"] is True:
            sparse_dim = self.dim["sparse"]
            results["sparse"] = []
            for sparse_vec in output["token_weights"]:
                indices = [int(k) for k in sparse_vec]
                values = list(sparse_vec.values())
                row_indices = [0] * len(indices)
                csr = csr_array((values, (row_indices, indices)), shape=(1, sparse_dim))
                results["sparse"].append(csr)
            results["sparse"] = stack_sparse_embeddings(results["sparse"]).tocsr()
        return results

    def encode_queries(self, queries: List[str]) -> Dict:
        return self._encode(queries)

    def encode_documents(self, documents: List[str]) -> Dict:
        return self._encode(documents)



from collections import defaultdict
from typing import List, Optional

import numpy as np
import struct

from pymilvus.model.base import BaseEmbeddingFunction
from pymilvus.model.utils import import_voyageai


class VoyageEmbeddingFunction(BaseEmbeddingFunction):
    def __init__(self,
                 model_name: str = "voyage-3-large",
                 api_key: Optional[str] = None,
                 embedding_type: Optional[str] = None,
                 truncate: Optional[bool] = None,
                 dimension: Optional[int] = None,
                 **kwargs):
        import_voyageai()
        import voyageai

        self.model_name = model_name
        self.truncate = truncate
        self._voyageai_model_meta_info = defaultdict(dict)
        self._voyageai_model_meta_info["voyage-3-large"]["dim"] = [1024, 256, 512, 2048]
        self._voyageai_model_meta_info["voyage-code-3"]["dim"] = [1024, 256, 512, 2048]
        self._voyageai_model_meta_info["voyage-3"]["dim"] = [1024]
        self._voyageai_model_meta_info["voyage-3-lite"]["dim"] = [512]
        self._voyageai_model_meta_info["voyage-finance-2"]["dim"] = [1024]
        self._voyageai_model_meta_info["voyage-multilingual-2"]["dim"] = [1024]
        self._voyageai_model_meta_info["voyage-law-2"]["dim"] = [1024]
        #old model
        self._voyageai_model_meta_info["voyage-large-2"]["dim"] = [1536]
        self._voyageai_model_meta_info["voyage-code-2"]["dim"] = [1536]
        self._voyageai_model_meta_info["voyage-2"]["dim"] = [1024]
        self._voyageai_model_meta_info["voyage-lite-02-instruct"]["dim"] = [1024]

        if dimension is not None and dimension not in self._voyageai_model_meta_info[self.model_name]["dim"]:
            raise ValueError(f"The provided dimension ({dimension}) is not supported by the selected model ({self.model_name}). "
                             "Leave this parameter empty to use the default dimension for the model. "
                             "Please check the supported dimensions here: https://docs.voyageai.com/docs/embeddings"
                             )

        if embedding_type == "int8" or embedding_type == "uint8":
            raise ValueError("Currently int8 or uint8 is not supported with PyMilvus model library.")

        if self.model_name in ['voyage-3-large', 'voyage-code-3']:
            if embedding_type is not None and embedding_type not in ['float', 'binary', 'ubinary']:
                raise ValueError(f"The provided embedding_type ({embedding_type}) is not supported by the selected model "
                                 f"({self.model_name}). Leave this parameter empty for the default embedding_type (float). "
                                 f"Please check the supported embedding_type values here: https://docs.voyageai.com/docs/embeddings")
        else:
            if embedding_type is not None and embedding_type != 'float':
                raise ValueError(f"The provided embedding_type ({embedding_type}) is not supported by the selected model "
                                 f"({self.model_name}). Leave this parameter empty for the default embedding_type (float). "
                                 f"Please check the supported embedding_type values here: https://docs.voyageai.com/docs/embeddings")

        self.embedding_type = embedding_type
        self.dimension = dimension
        self.client = voyageai.Client(api_key, **kwargs)

    @property
    def dim(self):
        if self.dimension is None:
            return self._voyageai_model_meta_info[self.model_name]["dim"][0]
        return self.dimension

    def encode_queries(self, queries: List[str]) -> List[np.array]:
        return self._call_voyage_api(queries, input_type="query")

    def encode_documents(self, documents: List[str]) -> List[np.array]:
        return self._call_voyage_api(documents, input_type="document")

    def __call__(self, texts: List[str]) -> List[np.array]:
        return self._call_voyage_api(texts)

    def _call_voyage_api(self, texts: List[str], input_type: Optional[str] = None):
        embeddings = self.client.embed(
            texts=texts,
            model=self.model_name,
            input_type=input_type,
            truncation=self.truncate,
            output_dtype=self.embedding_type,
            output_dimension=self.dim,
        ).embeddings

        if self.embedding_type is None or self.embedding_type == "float":
            results = [np.array(data, dtype=np.float32) for data in embeddings]
        elif self.embedding_type == "binary":
            results = [
                np.unpackbits((np.array(result, dtype=np.int16) + 128).astype(np.uint8)).astype(bool)
                for result in embeddings
            ]
        elif self.embedding_type == "ubinary":
            results = [np.unpackbits(np.array(result, dtype=np.uint8)).astype(bool) for result in embeddings]
        else:
            raise ValueError(f"The provided embedding_type ({self.embedding_type}) is not supported by the selected model "
                             f"({self.model_name}). Leave this parameter empty for the default embedding_type (float). "
                             f"Please check the supported embedding_type values here: https://docs.voyageai.com/docs/embeddings")
        return results



from typing import List
import numpy as np
import os
from collections import defaultdict

from pymilvus.model.base import BaseEmbeddingFunction
from pymilvus.model.utils import import_nomic


class NomicEmbeddingFunction(BaseEmbeddingFunction):
    def __init__(
        self,
        model_name: str = "nomic-embed-text-v1.5",
        task_type: str = "search_document",
        dimensions: int = 768,
        **kwargs,
    ):
        self._nomic_model_meta_info = defaultdict(dict)
        self._nomic_model_meta_info[model_name]["dim"] = dimensions  # set the dimension

        self.model_name = model_name
        self.task_type = task_type
        self.dimensionality = dimensions
        if "dimensionality" in kwargs: 
            self.dimensionality = kwargs["dimensionality"]
            kwargs.pop("dimensionality")

        self._encode_config = {
            "model": model_name,
            "task_type": task_type,
            "dimensionality": self.dimensionality,
            **kwargs,
        }

    def encode_queries(self, queries: List[str]) -> List[np.array]:
        return self._encode(queries, task_type="search_query")

    def encode_documents(self, documents: List[str]) -> List[np.array]:
        return self._encode(documents, task_type="search_document")

    @property
    def dim(self):
        return self._nomic_model_meta_info[self.model_name]["dim"]

    def __call__(self, texts: List[str]) -> List[np.array]:
        return self._encode(texts, task_type=self.task_type)

    def _encode_query(self, query: str) -> np.array:
        return self._encode([query], task_type="search_query")[0]

    def _encode_document(self, document: str) -> np.array:
        return self._encode([document], task_type="search_document")[0]

    def _call_nomic_api(self, texts: List[str], task_type: str):
        import_nomic()
        from nomic import embed

        embeddings_batch_response = embed.text(
            texts=texts,
            **self._encode_config
        )
        return [np.array(embedding) for embedding in embeddings_batch_response["embeddings"]]

    def _encode(self, texts: List[str], task_type: str):
        return self._call_nomic_api(texts, task_type)



from pymilvus.model.reranker.cohere import CohereRerankFunction
from pymilvus.model.reranker.bgereranker import BGERerankFunction
from pymilvus.model.reranker.voyageai import VoyageRerankFunction
from pymilvus.model.reranker.cross_encoder import CrossEncoderRerankFunction
from pymilvus.model.reranker.jinaai import JinaRerankFunction
from pymilvus.model.reranker.tei import TEIRerankFunction

__all__ = [
    "CohereRerankFunction",
    "BGERerankFunction",
    "VoyageRerankFunction",
    "CrossEncoderRerankFunction",
    "JinaRerankFunction",
    "TEIRerankFunction",
]



from typing import List, Optional
import os
import numpy as np
from collections import defaultdict

from pymilvus.model.base import BaseEmbeddingFunction
from pymilvus.model.utils import import_mistralai



class MistralAIEmbeddingFunction(BaseEmbeddingFunction):
    def __init__(
        self,
        api_key: str,
        model_name: str = "mistral-embed",
        **kwargs,
    ):
        import_mistralai()
        from mistralai import Mistral

        self._mistral_model_meta_info = defaultdict(dict)
        self._mistral_model_meta_info[model_name]["dim"] = 1024  # fixed dimension

        if api_key is None:
            if "MISTRALAI_API_KEY" in os.environ and os.environ["MISTRALAI_API_KEY"]:
                self.api_key = os.environ["MISTRALAI_API_KEY"]
            else:
                error_message = (
                    "Did not find api_key, please add an environment variable"
                    " `MISTRALAI_API_KEY` which contains it, or pass"
                    "  `api_key` as a named parameter."
                )
                raise ValueError(error_message)
        else:
            self.api_key = api_key
        self.model_name = model_name
        self.client = Mistral(api_key=api_key)
        self._encode_config = {"model": model_name, **kwargs}

    def encode_queries(self, queries: List[str]) -> List[np.array]:
        return self._encode(queries)

    def encode_documents(self, documents: List[str]) -> List[np.array]:
        return self._encode(documents)

    @property
    def dim(self):
        return self._mistral_model_meta_info[self.model_name]["dim"]

    def __call__(self, texts: List[str]) -> List[np.array]:
        return self._encode(texts)

    def _encode_query(self, query: str) -> np.array:
        return self._encode([query])[0]

    def _encode_document(self, document: str) -> np.array:
        return self._encode([document])[0]

    def _call_mistral_api(self, texts: List[str]):
        embeddings_batch_response = self.client.embeddings.create(
            inputs=texts,
            **self._encode_config
        )
        return [np.array(data.embedding) for data in embeddings_batch_response.data]

    def _encode(self, texts: List[str]):
        return self._call_mistral_api(texts)



"""
This file incorporates components from the 'rank_bm25' project by Dorian Brown:
https://github.com/dorianbrown/rank_bm25
Specifically, the rank_bm25.py file.

The incorporated components are licensed under the Apache License, Version 2.0 (the "License");
you may not use these components except in compliance with the License.
You may obtain a copy of the License at http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

import json
import logging
import math
from collections import defaultdict
from multiprocessing import Pool
from pathlib import Path
from typing import Dict, List, Optional

import requests
from scipy.sparse import csr_array, vstack
import numpy as np

from pymilvus.model.base import BaseEmbeddingFunction
from pymilvus.model.sparse.bm25.tokenizers import Analyzer, build_default_analyzer

logger = logging.getLogger(__name__)
logger.setLevel(logging.DEBUG)
console_handler = logging.StreamHandler()
console_handler.setLevel(logging.INFO)
logger.addHandler(console_handler)


class BM25EmbeddingFunction(BaseEmbeddingFunction):
    def __init__(
        self,
        analyzer: Analyzer = None,
        corpus: Optional[List] = None,
        k1: float = 1.5,
        b: float = 0.75,
        epsilon: float = 0.25,
        num_workers: int = 1,
    ):
        if analyzer is None:
            analyzer = build_default_analyzer(language="en")
        self.analyzer = analyzer
        self.corpus_size = 0
        self.avgdl = 0
        self.idf = {}
        self.k1 = k1
        self.b = b
        self.epsilon = epsilon
        self.num_workers = num_workers

        if analyzer and corpus is not None:
            self.fit(corpus)

    def _calc_term_indices(self):
        for index, word in enumerate(self.idf):
            self.idf[word][1] = index

    def _compute_statistics(self, corpus: List[str]):
        term_document_frequencies = defaultdict(int)
        total_word_count = 0
        for document in corpus:
            total_word_count += len(document)

            frequencies = defaultdict(int)
            for word in document:
                frequencies[word] += 1

            for word, _ in frequencies.items():
                term_document_frequencies[word] += 1
            self.corpus_size += 1
        self.avgdl = total_word_count / self.corpus_size
        return term_document_frequencies

    def _tokenize_corpus(self, corpus: List[str]):
        if self.num_workers == 1:
            return [self.analyzer(text) for text in corpus]
        pool = Pool(self.num_workers)
        return pool.map(self.analyzer, corpus)

    def _calc_idf(self, term_document_frequencies: Dict):
        # collect idf sum to calculate an average idf for epsilon value
        idf_sum = 0
        # collect words with negative idf to set them a special epsilon value.
        # idf can be negative if word is contained in more than half of documents
        negative_idfs = []
        for word, freq in term_document_frequencies.items():
            idf = math.log(self.corpus_size - freq + 0.5) - math.log(freq + 0.5)
            if word not in self.idf:
                self.idf[word] = [0.0, 0]
            self.idf[word][0] = idf
            idf_sum += idf
            if idf < 0:
                negative_idfs.append(word)
        self.average_idf = idf_sum / len(self.idf)

        eps = self.epsilon * self.average_idf
        for word in negative_idfs:
            self.idf[word][0] = eps

    def _rebuild(self, corpus: List[str]):
        self._clear()
        corpus = self._tokenize_corpus(corpus)
        term_document_frequencies = self._compute_statistics(corpus)
        self._calc_idf(term_document_frequencies)
        self._calc_term_indices()

    def _clear(self):
        self.corpus_size = 0
        # idf records the (value, index)
        self.idf = defaultdict(list)

    @property
    def dim(self):
        return len(self.idf)

    def fit(self, corpus: List[str]):
        self._rebuild(corpus)

    def _encode_query(self, query: str) -> csr_array:
        terms = self.analyzer(query)
        values, rows, cols = [], [], []
        for term in terms:
            if term in self.idf:
                values.append(self.idf[term][0])
                rows.append(0)
                cols.append(self.idf[term][1])
        return csr_array((values, (rows, cols)), shape=(1, len(self.idf))).astype(np.float32)

    def _encode_document(self, doc: str) -> csr_array:
        terms = self.analyzer(doc)
        frequencies = defaultdict(int)
        doc_len = len(terms)
        term_set = set()
        for term in terms:
            frequencies[term] += 1
            term_set.add(term)
        values, rows, cols = [], [], []
        for term in term_set:
            if term in self.idf:
                term_freq = frequencies[term]
                value = (
                    term_freq
                    * (self.k1 + 1)
                    / (term_freq + self.k1 * (1 - self.b + self.b * doc_len / self.avgdl))
                )
                rows.append(0)
                cols.append(self.idf[term][1])
                values.append(value)
        return csr_array((values, (rows, cols)), shape=(1, len(self.idf))).astype(np.float32)

    def encode_queries(self, queries: List[str]) -> csr_array:
        sparse_embs = [self._encode_query(query) for query in queries]
        return vstack(sparse_embs).tocsr()

    def __call__(self, texts: List[str]) -> csr_array:
        error_message = "Unsupported function called, please check the documentation of 'BM25EmbeddingFunction'."
        raise ValueError(error_message)

    def encode_documents(self, documents: List[str]) -> csr_array:
        sparse_embs = [self._encode_document(document) for document in documents]
        return vstack(sparse_embs).tocsr()

    def save(self, path: str):
        bm25_params = {}
        bm25_params["version"] = "v1"
        bm25_params["corpus_size"] = self.corpus_size
        bm25_params["avgdl"] = self.avgdl
        bm25_params["idf_word"] = [None for _ in range(len(self.idf))]
        bm25_params["idf_value"] = [None for _ in range(len(self.idf))]
        for word, values in self.idf.items():
            bm25_params["idf_word"][values[1]] = word
            bm25_params["idf_value"][values[1]] = values[0]

        bm25_params["k1"] = self.k1
        bm25_params["b"] = self.b
        bm25_params["epsilon"] = self.epsilon

        with Path(path).open("w") as json_file:
            json.dump(bm25_params, json_file)

    def load(self, path: Optional[str] = None):
        default_meta_filename = "bm25_msmarco_v1.json"
        default_meta_url = "https://github.com/milvus-io/pymilvus-assets/releases/download/v0.1-bm25v1/bm25_msmarco_v1.json"
        if path is None:
            logger.info(f"path is None, using default {default_meta_filename}.")
            if not Path(default_meta_filename).exists():
                try:
                    logger.info(
                        f"{default_meta_filename} not found, start downloading from {default_meta_url} to ./{default_meta_filename}."
                    )
                    response = requests.get(default_meta_url, timeout=30)
                    response.raise_for_status()
                    with Path(default_meta_filename).open("wb") as f:
                        f.write(response.content)
                    logger.info(f"{default_meta_filename} has been downloaded successfully.")
                except requests.exceptions.RequestException as e:
                    error_message = f"Failed to download the file: {e}"
                    raise RuntimeError(error_message) from e
            path = default_meta_filename
        try:
            with Path(path).open() as json_file:
                bm25_params = json.load(json_file)
        except OSError as e:
            error_message = f"Error opening file {path}: {e}"
            raise RuntimeError(error_message) from e
        self.corpus_size = bm25_params["corpus_size"]
        self.avgdl = bm25_params["avgdl"]
        self.idf = {}
        for i in range(len(bm25_params["idf_word"])):
            self.idf[bm25_params["idf_word"][i]] = [bm25_params["idf_value"][i], i]
        self.k1 = bm25_params["k1"]
        self.b = bm25_params["b"]
        self.epsilon = bm25_params["epsilon"]



__all__ = [
    "import_openai",
    "import_sentence_transformers",
    "import_FlagEmbedding",
    "import_nltk",
    "import_transformers",
    "import_jieba",
    "import_konlpy",
    "import_mecab",
    "import_scipy",
    "import_protobuf",
    "import_unidic_lite",
    "import_cohere",
    "import_voyageai",
    "import_torch",
    "import_huggingface_hub",
    "import_mistralai",
    "import_nomic",
    "import_instructor",
    "import_datasets",
    "import_model2vec",
    ]

import importlib.util
from typing import Optional

from pymilvus.model.utils.dependency_control import prompt_install

def import_openai():
    _check_library("openai", package="openai>=1.12.0")

def import_sentence_transformers():
    _check_library("sentence_transformers", package="sentence-transformers")

def import_FlagEmbedding():
    _check_library("peft", package="peft")
    _check_library("FlagEmbedding", package="FlagEmbedding>=1.3.3")

def import_nltk():
    _check_library("nltk", package="nltk>=3.9.1")

def import_transformers():
    _check_library("transformers", package="transformers>=4.36.0")

def import_jieba():
    _check_library("jieba", package="jieba")

def import_konlpy():
    _check_library("konlpy", package="konlpy")

def import_mecab():
    _check_library("konlpy", package="mecab-python3")

def import_scipy():
    _check_library("scipy", package="scipy>=1.10.0")

def import_protobuf():
    _check_library("protobuf", package="protobuf==3.20.2")

def import_unidic_lite():
    _check_library("unidic-lite", package="unidic-lite")

def import_cohere():
    _check_library("cohere", "cohere>=5.10.0")

def import_voyageai():
    _check_library("voyageai", "voyageai>=0.2.0")

def import_torch():
    _check_library("torch", "torch")

def import_huggingface_hub():
    _check_library("huggingface_hub", package="huggingface-hub")

def import_mistralai():
    _check_library("mistralai", package="mistralai")

def import_nomic():
    _check_library("nomic", package="nomic")

def import_instructor():
    _check_library("InstructorEmbedding", package="InstructorEmbedding")

def import_datasets():
    _check_library("datasets", package="datasets")

def import_model2vec():
    _check_library("model2vec", package="model2vec")

def _check_library(libname: str, prompt: bool = True, package: Optional[str] = None):
    is_avail = False
    if importlib.util.find_spec(libname):
        is_avail = True
    if not is_avail and prompt:
        prompt_install(package if package else libname)
    return is_avail



"""
The following code is adapted from/inspired by the 'neural-cherche' project:
https://github.com/raphaelsty/neural-cherche
Specifically, neural-cherche/neural_cherche/models/splade.py

MIT License

Copyright (c) 2023 Raphael Sourty

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
"""

import logging
from typing import List, Optional

from scipy.sparse import csr_array

from pymilvus.model.base import BaseEmbeddingFunction


logger = logging.getLogger(__name__)
logger.setLevel(logging.DEBUG)


class SpladeEmbeddingFunction(BaseEmbeddingFunction):
    model_name: str

    def __init__(
        self,
        model_name: str = "naver/splade-cocondenser-ensembledistil",
        batch_size: int = 32,
        query_instruction: str = "",
        doc_instruction: str = "",
        device: Optional[str] = "cpu",
        k_tokens_query: Optional[int] = None,
        k_tokens_document: Optional[int] = None,
        **kwargs,
    ):
        from .splade_embedding.splade_impl import _SpladeImplementation
        self.model_name = model_name

        _model_config = dict(
            {"model_name_or_path": model_name, "batch_size": batch_size, "device": device},
            **kwargs,
        )
        self._model_config = _model_config
        self.model = _SpladeImplementation(**self._model_config)
        self.device = device
        self.k_tokens_query = k_tokens_query
        self.k_tokens_document = k_tokens_document
        self.query_instruction = query_instruction
        self.doc_instruction = doc_instruction

    def __call__(self, texts: List[str]) -> csr_array:
        return self._encode(texts, None)

    def encode_documents(self, documents: List[str]) -> csr_array:
        return self._encode(
            [self.doc_instruction + document for document in documents], self.k_tokens_document,
        )

    def _encode(self, texts: List[str], k_tokens: int) -> csr_array:
        return self.model.forward(texts, k_tokens=k_tokens)

    def encode_queries(self, queries: List[str]) -> csr_array:
        return self._encode(
            [self.query_instruction + query for query in queries], self.k_tokens_query,
        )

    @property
    def dim(self) -> int:
        return len(self.model.tokenizer)

    def _encode_query(self, query: str) -> csr_array:
        return self.model.forward([self.query_instruction + query], k_tokens=self.k_tokens_query)[0]

    def _encode_document(self, document: str) -> csr_array:
        return self.model.forward(
            [self.doc_instruction + document], k_tokens=self.k_tokens_document
        )[0]




from typing import List

import numpy as np

from pymilvus.model.base import BaseEmbeddingFunction
from pymilvus.model.utils import import_sentence_transformers


class SentenceTransformerEmbeddingFunction(BaseEmbeddingFunction):
    def __init__(
        self,
        model_name: str = "all-MiniLM-L6-v2",
        batch_size: int = 32,
        query_instruction: str = "",
        doc_instruction: str = "",
        device: str = "cpu",
        normalize_embeddings: bool = True,
        **kwargs,
    ):
        import_sentence_transformers()
        from sentence_transformers import SentenceTransformer
        self.model_name = model_name
        self.query_instruction = query_instruction
        self.doc_instruction = doc_instruction
        self.batch_size = batch_size
        self.normalize_embeddings = normalize_embeddings

        _model_config = dict({"model_name_or_path": model_name, "device": device}, **kwargs)
        self.model = SentenceTransformer(**_model_config)

    def __call__(self, texts: List[str]) -> List[np.array]:
        return self._encode(texts)

    def _encode(self, texts: List[str]) -> List[np.array]:
        embs = self.model.encode(
            texts, 
            batch_size=self.batch_size,
            show_progress_bar=False,
            convert_to_numpy=True,
            normalize_embeddings=self.normalize_embeddings
        )
        return list(embs)

    @property
    def dim(self):
        return self.model.get_sentence_embedding_dimension()

    def encode_queries(self, queries: List[str]) -> List[np.array]:
        instructed_queries = [self.query_instruction + query for query in queries]
        return self._encode(instructed_queries)

    def encode_documents(self, documents: List[str]) -> List[np.array]:
        instructed_documents = [self.doc_instruction + document for document in documents]
        return self._encode(instructed_documents)

    def _encode_query(self, query: str) -> np.array:
        instructed_query = self.query_instruction + query
        embs = self.model.encode(
            sentences=[instructed_query],
            batch_size=1,
            show_progress_bar=False,
            convert_to_numpy=True,
            normalize_embeddings=self.normalize_embeddings,
        )
        return embs[0]

    def _encode_document(self, document: str) -> np.array:
        instructed_document = self.doc_instruction + document
        embs = self.model.encode(
            sentences=[instructed_document],
            batch_size=1,
            show_progress_bar=False,
            convert_to_numpy=True,
            normalize_embeddings=self.normalize_embeddings,
        )
        return embs[0]
